@@ -25,7 +25,10 @@
 
 #include "giggle-git.h"
 #include "giggle-file-list.h"
+#include "giggle-diff-window.h"
 #include "giggle-git-ignore.h"
+#include "giggle-git-list-files.h"
+#include "giggle-git-add.h"
 #include "giggle-git-diff-tree.h"
 #include "giggle-revision.h"
 
@@ -43,8 +46,10 @@ struct GiggleFileListPriv {
 
 	GiggleJob    *job;
 
-	gboolean      show_all : 1;
-	gboolean      compact_mode : 1;
+	GtkWidget    *diff_window;
+
+	guint         show_all : 1;
+	guint         compact_mode : 1;
 };
 
 static void       file_list_finalize           (GObject        *object);
@@ -62,6 +67,8 @@ static gboolean   file_list_button_press       (GtkWidget      *widget,
 static void       file_list_directory_changed  (GObject        *object,
 						GParamSpec     *pspec,
 						gpointer        user_data);
+static void       file_list_managed_files_changed (GiggleFileList *list);
+
 static void       file_list_populate           (GiggleFileList *list);
 static void       file_list_populate_dir       (GiggleFileList *list,
 						const gchar    *directory,
@@ -76,17 +83,25 @@ static gboolean   file_list_search_equal_func   (GtkTreeModel   *model,
 						 const gchar    *key,
 						 GtkTreeIter    *iter,
 						 gpointer        search_data);
+static gint       file_list_compare_func        (GtkTreeModel   *model,
+						 GtkTreeIter    *a,
+						 GtkTreeIter    *b,
+						 gpointer        user_data);
 
-static gboolean   file_list_get_name_and_ignore_for_iter (GiggleFileList   *list,
+static gboolean   file_list_get_path_and_ignore_for_iter (GiggleFileList   *list,
 							  GtkTreeIter      *iter,
 							  gchar           **name,
 							  GiggleGitIgnore **git_ignore);
 
-static void       file_list_add_file            (GtkWidget        *widget,
+static void       file_list_diff_file           (GtkAction        *action,
 						 GiggleFileList   *list);
-static void       file_list_remove_file         (GtkWidget        *widget,
+static void       file_list_add_file            (GtkAction        *action,
 						 GiggleFileList   *list);
-static void       file_list_toggle_show_all     (GtkWidget        *widget,
+static void       file_list_ignore_file         (GtkAction        *action,
+						 GiggleFileList   *list);
+static void       file_list_unignore_file       (GtkAction        *action,
+						 GiggleFileList   *list);
+static void       file_list_toggle_show_all     (GtkAction        *action,
 						 GiggleFileList   *list);
 
 static void       file_list_cell_data_sensitive_func  (GtkCellLayout   *cell_layout,
@@ -100,7 +115,6 @@ static void       file_list_cell_data_background_func (GtkCellLayout   *cell_lay
 						       GtkTreeIter     *iter,
 						       gpointer         data);
 
-
 G_DEFINE_TYPE (GiggleFileList, giggle_file_list, GTK_TYPE_TREE_VIEW)
 
 #define GET_PRIV(obj) (G_TYPE_INSTANCE_GET_PRIVATE ((obj), GIGGLE_TYPE_FILE_LIST, GiggleFileListPriv))
@@ -110,6 +124,7 @@ enum {
 	COL_REL_PATH,
 	COL_PIXBUF,
 	COL_GIT_IGNORE,
+	COL_MANAGED, /* File managed by git */
 	COL_HIGHLIGHT,
 	LAST_COL
 };
@@ -120,20 +135,33 @@ enum {
 	PROP_COMPACT_MODE,
 };
 
-GtkActionEntry menu_items [] = {
-	{ "Add",    GTK_STOCK_ADD,    N_("_Add to .gitignore"),      NULL, NULL, G_CALLBACK (file_list_add_file) },
-	{ "Remove", GTK_STOCK_REMOVE, N_("_Remove from .gitignore"), NULL, NULL, G_CALLBACK (file_list_remove_file) },
+enum {
+	PATH_SELECTED,
+	LAST_SIGNAL
 };
 
-GtkToggleActionEntry toggle_menu_items [] = {
+static guint signals[LAST_SIGNAL] = { 0, };
+
+static GtkActionEntry menu_items [] = {
+	{ "Diff",     NULL,             N_("_Diff"),                   NULL, NULL, G_CALLBACK (file_list_diff_file) },
+	{ "AddFile",  NULL,             N_("A_dd file to repository"), NULL, NULL, G_CALLBACK (file_list_add_file) },
+	{ "Ignore",   GTK_STOCK_ADD,    N_("_Add to .gitignore"),      NULL, NULL, G_CALLBACK (file_list_ignore_file) },
+	{ "Unignore", GTK_STOCK_REMOVE, N_("_Remove from .gitignore"), NULL, NULL, G_CALLBACK (file_list_unignore_file) },
+};
+
+static GtkToggleActionEntry toggle_menu_items [] = {
 	{ "ShowAll", NULL, N_("_Show all files"), NULL, NULL, G_CALLBACK (file_list_toggle_show_all), FALSE },
 };
 
-const gchar *ui_description =
+static const gchar *ui_description =
 	"<ui>"
 	"  <popup name='PopupMenu'>"
-	"    <menuitem action='Add'/>"
-	"    <menuitem action='Remove'/>"
+	"    <menuitem action='Diff'/>"
+	"    <separator/>"
+	"    <menuitem action='AddFile'/>"
+	"    <separator/>"
+	"    <menuitem action='Ignore'/>"
+	"    <menuitem action='Unignore'/>"
 	"    <separator/>"
 	"    <menuitem action='ShowAll'/>"
 	"  </popup>"
@@ -167,6 +195,16 @@ giggle_file_list_class_init (GiggleFileListClass *class)
 							       FALSE,
 							       G_PARAM_READWRITE));
 
+	signals[PATH_SELECTED] =
+		g_signal_new ("path-selected",
+			      G_OBJECT_CLASS_TYPE (object_class),
+			      G_SIGNAL_RUN_LAST,
+			      G_STRUCT_OFFSET (GiggleFileListClass, path_selected),
+			      NULL, NULL,
+			      g_cclosure_marshal_VOID__STRING,
+			      G_TYPE_NONE,
+			      1, G_TYPE_STRING);
+
 	g_type_class_add_private (object_class, sizeof (GiggleFileListPriv));
 }
 
@@ -182,17 +220,25 @@ giggle_file_list_init (GiggleFileList *list)
 	priv = GET_PRIV (list);
 
 	priv->git = giggle_git_get ();
-	g_signal_connect (G_OBJECT (priv->git), "notify::project-dir",
+	g_signal_connect (priv->git, "notify::project-dir",
 			  G_CALLBACK (file_list_directory_changed), list);
+	g_signal_connect_swapped (priv->git, "notify::git-dir",
+				  G_CALLBACK (file_list_managed_files_changed), list);
 
 	priv->icon_theme = gtk_icon_theme_get_default ();
 
-	priv->store = gtk_tree_store_new (LAST_COL, G_TYPE_STRING, G_TYPE_STRING, GDK_TYPE_PIXBUF, GIGGLE_TYPE_GIT_IGNORE, G_TYPE_BOOLEAN);
+	priv->store = gtk_tree_store_new (LAST_COL, G_TYPE_STRING, G_TYPE_STRING, GDK_TYPE_PIXBUF,
+					  GIGGLE_TYPE_GIT_IGNORE, G_TYPE_BOOLEAN, G_TYPE_BOOLEAN, G_TYPE_BOOLEAN);
 	priv->filter_model = gtk_tree_model_filter_new (GTK_TREE_MODEL (priv->store), NULL);
 
 	gtk_tree_model_filter_set_visible_func (GTK_TREE_MODEL_FILTER (priv->filter_model),
 						file_list_filter_func, list, NULL);
 
+	gtk_tree_sortable_set_sort_func (GTK_TREE_SORTABLE (priv->store),
+					 COL_NAME,
+					 file_list_compare_func,
+					 list, NULL);
+
 	gtk_tree_sortable_set_sort_column_id (GTK_TREE_SORTABLE (priv->store),
 					      COL_NAME, GTK_SORT_ASCENDING);
 
@@ -216,19 +262,16 @@ giggle_file_list_init (GiggleFileList *list)
 	gtk_cell_layout_set_cell_data_func (GTK_CELL_LAYOUT (column), renderer,
 					    file_list_cell_data_background_func,
 					    list, NULL);
-					    
 
 	renderer = gtk_cell_renderer_text_new ();
-	gtk_cell_layout_pack_start (GTK_CELL_LAYOUT (column), renderer, FALSE);
+	g_object_set (renderer, "ellipsize", PANGO_ELLIPSIZE_END, NULL);
+	gtk_cell_layout_pack_start (GTK_CELL_LAYOUT (column), renderer, TRUE);
 	gtk_cell_layout_set_attributes (GTK_CELL_LAYOUT (column), renderer,
 					"text", COL_NAME,
 					NULL);
 	gtk_cell_layout_set_cell_data_func (GTK_CELL_LAYOUT (column), renderer,
 					    file_list_cell_data_sensitive_func,
 					    list, NULL);
-	gtk_cell_layout_set_cell_data_func (GTK_CELL_LAYOUT (column), renderer,
-					    file_list_cell_data_background_func,
-					    list, NULL);
 
 	gtk_tree_view_append_column (GTK_TREE_VIEW (list), column);
 	file_list_populate (list);
@@ -256,6 +299,14 @@ giggle_file_list_init (GiggleFileList *list)
 			     "  GtkTreeView::vertical-separator = 0"
 			     "}"
 			     "widget \"*.file-list\" style \"file-list-compact-style\"");
+
+	/* create diff window */
+	priv->diff_window = giggle_diff_window_new ();
+
+	g_signal_connect (priv->diff_window, "delete-event",
+			  G_CALLBACK (gtk_widget_hide_on_delete), NULL);
+	g_signal_connect_after (priv->diff_window, "response",
+				G_CALLBACK (gtk_widget_hide), NULL);
 }
 
 static void
@@ -334,50 +385,88 @@ file_list_button_press (GtkWidget      *widget,
 {
 	GiggleFileList     *list;
 	GiggleFileListPriv *priv;
-	gboolean            add, remove;
-	gchar              *name;
+	gboolean            add, ignore, unignore;
+	gchar              *file_path;
 	GiggleGitIgnore    *git_ignore;
 	GtkAction          *action;
 	GtkTreeSelection   *selection;
 	GtkTreeModel       *model;
 	GList              *rows, *l;
 	GtkTreeIter         iter;
+	GtkTreePath        *path;
 
 	list = GIGGLE_FILE_LIST (widget);
 	priv = GET_PRIV (list);
-	add = remove = FALSE;
-
-	GTK_WIDGET_CLASS (giggle_file_list_parent_class)->button_press_event (widget, event);
+	ignore = unignore = add = FALSE;
 
 	if (event->button == 3) {
 		selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (list));
+
+		if (!gtk_tree_view_get_path_at_pos (GTK_TREE_VIEW (widget), event->x, event->y,
+						    &path, NULL, NULL, NULL)) {
+			return TRUE;
+		}
+
+		if (event->state != 0) {
+			/* we want shift+click and ctrl+click behave as always here */
+			GTK_WIDGET_CLASS (giggle_file_list_parent_class)->button_press_event (widget, event);
+		} else if (!gtk_tree_selection_path_is_selected (selection, path)) {
+			gtk_tree_selection_unselect_all (selection);
+			gtk_tree_selection_select_path (selection, path);
+		}
+
+		gtk_tree_path_free (path);
 		rows = gtk_tree_selection_get_selected_rows (selection, &model);
 
 		for (l = rows; l; l = l->next) {
 			gtk_tree_model_get_iter (model, &iter, l->data);
 
-			if (file_list_get_name_and_ignore_for_iter (list, &iter, &name, &git_ignore)) {
-				if (giggle_git_ignore_name_matches (git_ignore, name)) {
-					remove = TRUE;
+			gtk_tree_model_get (model, &iter,
+					    COL_MANAGED, &add,
+					    -1);
+
+			if (file_list_get_path_and_ignore_for_iter (list, &iter, &file_path, &git_ignore)) {
+				if (giggle_git_ignore_path_matches (git_ignore, file_path)) {
+					unignore = TRUE;
 				} else {
-					add = TRUE;
+					ignore = TRUE;
 				}
 
 				g_object_unref (git_ignore);
-				g_free (name);
+				g_free (file_path);
 			}
 		}
 
-		action = gtk_ui_manager_get_action (priv->ui_manager, "/ui/PopupMenu/Add");
+		action = gtk_ui_manager_get_action (priv->ui_manager, "/ui/PopupMenu/AddFile");
 		gtk_action_set_sensitive (action, add);
-		action = gtk_ui_manager_get_action (priv->ui_manager, "/ui/PopupMenu/Remove");
-		gtk_action_set_sensitive (action, remove);
+		action = gtk_ui_manager_get_action (priv->ui_manager, "/ui/PopupMenu/Ignore");
+		gtk_action_set_sensitive (action, ignore);
+		action = gtk_ui_manager_get_action (priv->ui_manager, "/ui/PopupMenu/Unignore");
+		gtk_action_set_sensitive (action, unignore);
 		
 		gtk_menu_popup (GTK_MENU (priv->popup), NULL, NULL,
 				NULL, NULL, event->button, event->time);
 
 		g_list_foreach (rows, (GFunc) gtk_tree_path_free, NULL);
 		g_list_free (rows);
+	} else {
+		GTK_WIDGET_CLASS (giggle_file_list_parent_class)->button_press_event (widget, event);
+
+		if (event->button == 1 &&
+		    event->state == 0 &&
+		    event->type == GDK_2BUTTON_PRESS) {
+			selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (list));
+			rows = gtk_tree_selection_get_selected_rows (selection, &model);
+
+			g_assert (rows != NULL);
+
+			/* there should be just one item selected */
+			gtk_tree_model_get_iter (model, &iter, rows->data);
+
+			file_list_get_path_and_ignore_for_iter (list, &iter, &file_path, NULL);
+			g_signal_emit (widget, signals[PATH_SELECTED], 0, file_path);
+			g_free (file_path);
+		}
 	}
 
 	return TRUE;
@@ -398,6 +487,135 @@ file_list_directory_changed (GObject    *object,
 	file_list_populate (list);
 }
 
+static void
+file_list_update_managed (GiggleFileList *file_list,
+			  GtkTreeIter    *parent,
+			  const gchar    *parent_path,
+			  GList          *files)
+{
+	GiggleFileListPriv *priv;
+	GtkTreeIter         iter;
+	gboolean            valid;
+	GiggleGitIgnore    *git_ignore;
+	gchar              *name, *path;
+	gboolean            managed;
+
+	priv = GET_PRIV (file_list);
+
+	if (parent) {
+		valid = gtk_tree_model_iter_children (GTK_TREE_MODEL (priv->store),
+						      &iter, parent);
+	} else {
+		valid = gtk_tree_model_get_iter_first (GTK_TREE_MODEL (priv->store), &iter);
+	}
+
+	while (valid) {
+		gtk_tree_model_get (GTK_TREE_MODEL (priv->store), &iter,
+				    COL_NAME, &name,
+				    COL_GIT_IGNORE, &git_ignore,
+				    -1);
+
+		if (parent_path) {
+			path = g_build_filename (parent_path, name, NULL);
+			managed = (g_list_find_custom (files, path, (GCompareFunc) strcmp) != NULL);
+		} else {
+			/* we don't want the project basename included */
+			path = g_strdup ("");
+			managed = FALSE;
+		}
+
+		gtk_tree_store_set (priv->store, &iter,
+				    COL_MANAGED, managed,
+				    -1);
+
+		if (git_ignore) {
+			/* it's a directory */
+			file_list_update_managed (file_list, &iter, path, files);
+			g_object_unref (git_ignore);
+		}
+
+		g_free (path);
+		g_free (name);
+		valid = gtk_tree_model_iter_next (GTK_TREE_MODEL (priv->store), &iter);
+	}
+}
+
+static void
+file_list_files_callback (GiggleGit *git,
+			  GiggleJob *job,
+			  GError    *error,
+			  gpointer   user_data)
+{
+	GiggleFileList     *list;
+	GiggleFileListPriv *priv;
+	GList              *files;
+
+	list = GIGGLE_FILE_LIST (user_data);
+	priv = GET_PRIV (list);
+
+	if (error) {
+		GtkWidget *dialog;
+
+		dialog = gtk_message_dialog_new (GTK_WINDOW (gtk_widget_get_toplevel (GTK_WIDGET (list))),
+						 GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
+						 GTK_MESSAGE_ERROR,
+						 GTK_BUTTONS_OK,
+						 _("An error ocurred when retrieving the file list:\n%s"),
+						 error->message);
+
+		gtk_dialog_run (GTK_DIALOG (dialog));
+		gtk_widget_destroy (dialog);
+	} else {
+		files = giggle_git_list_files_get_files (GIGGLE_GIT_LIST_FILES (priv->job));
+		file_list_update_managed (list, NULL, NULL, files);
+	}
+
+	g_object_unref (priv->job);
+	priv->job = NULL;
+}
+
+static void
+file_list_managed_files_changed (GiggleFileList *list)
+{
+	GiggleFileListPriv *priv;
+
+	priv = GET_PRIV (list);
+
+	file_list_update_managed (list, NULL, NULL, NULL);
+
+	if (priv->job) {
+		giggle_git_cancel_job (priv->git, priv->job);
+		g_object_unref (priv->job);
+		priv->job = NULL;
+	}
+
+	priv->job = giggle_git_list_files_new ();
+
+	giggle_git_run_job (priv->git,
+			    priv->job,
+			    file_list_files_callback,
+			    list);
+}
+
+static void
+file_list_diff_file (GtkAction      *action,
+		     GiggleFileList *list)
+{
+	GiggleFileListPriv *priv;
+	GtkWidget          *toplevel;
+	GList              *files;
+
+	priv = GET_PRIV (list);
+
+	files = giggle_file_list_get_selection (list);
+	giggle_diff_window_set_files (GIGGLE_DIFF_WINDOW (priv->diff_window), files);
+
+	toplevel = gtk_widget_get_toplevel (GTK_WIDGET (list));
+	gtk_window_set_transient_for (GTK_WINDOW (priv->diff_window),
+				      GTK_WINDOW (toplevel));
+	gtk_widget_show (priv->diff_window);
+}
+
 static void
 file_list_add_element (GiggleFileList *list,
 		       const gchar    *directory,
@@ -428,7 +646,7 @@ file_list_add_element (GiggleFileList *list,
 		git_ignore = giggle_git_ignore_new (full_path);
 	} else {
 		pixbuf = gtk_icon_theme_load_icon (priv->icon_theme,
-						   "gnome-mime-text", 16, 0, NULL);;
+						   "text-x-generic", 16, 0, NULL);;
 	}
 
 	gtk_tree_store_set (priv->store, &iter,
@@ -493,10 +711,69 @@ file_list_populate (GiggleFileList *list)
 	gtk_tree_path_free (path);
 }
 
+static void
+file_list_add_file_callback (GiggleGit *git,
+			     GiggleJob *job,
+			     GError    *error,
+			     gpointer   user_data)
+{
+	GiggleFileList     *list;
+	GiggleFileListPriv *priv;
+
+	list = GIGGLE_FILE_LIST (user_data);
+	priv = GET_PRIV (list);
+
+	if (error) {
+		GtkWidget *dialog;
+
+		dialog = gtk_message_dialog_new (GTK_WINDOW (gtk_widget_get_toplevel (GTK_WIDGET (list))),
+						 GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
+						 GTK_MESSAGE_ERROR,
+						 GTK_BUTTONS_OK,
+						 _("An error ocurred when adding a file to git:\n%s"),
+						 error->message);
+
+		gtk_dialog_run (GTK_DIALOG (dialog));
+		gtk_widget_destroy (dialog);
+
+		g_object_unref (priv->job);
+		priv->job = NULL;
+	} else {
+		g_object_unref (priv->job);
+		priv->job = NULL;
+
+		file_list_managed_files_changed (list);
+	}
+}
+
+static void
+file_list_add_file (GtkAction      *action,
+		    GiggleFileList *list)
+{
+	GiggleFileListPriv *priv;
+
+	priv = GET_PRIV (list);
+
+	if (priv->job) {
+		giggle_git_cancel_job (priv->git, priv->job);
+		g_object_unref (priv->job);
+		priv->job = NULL;
+	}
+
+	priv->job = giggle_git_add_new ();
+	giggle_git_add_set_files (GIGGLE_GIT_ADD (priv->job),
+				  giggle_file_list_get_selection (list));
+
+	giggle_git_run_job (priv->git,
+			    priv->job,
+			    file_list_add_file_callback,
+			    list);
+}
+
 static gboolean
-file_list_ignore_file (GtkTreeModel *model,
-		       GtkTreeIter  *file_iter,
-		       const gchar  *name)
+file_list_get_ignore_file (GtkTreeModel *model,
+			   GtkTreeIter  *file_iter,
+			   const gchar  *path)
 {
 	GiggleGitIgnore *git_ignore;
 	GtkTreeIter      iter, parent;
@@ -510,7 +787,7 @@ file_list_ignore_file (GtkTreeModel *model,
 				    -1);
 
 		if (git_ignore) {
-			matches = giggle_git_ignore_name_matches (git_ignore, name);
+			matches = giggle_git_ignore_path_matches (git_ignore, path);
 			g_object_unref (git_ignore);
 		}
 
@@ -528,35 +805,35 @@ file_list_filter_func (GtkTreeModel   *model,
 {
 	GiggleFileList     *list;
 	GiggleFileListPriv *priv;
-	gchar              *name;
+	gchar              *path;
 	gboolean            retval = TRUE;
 
 	list = GIGGLE_FILE_LIST (user_data);
 	priv = GET_PRIV (list);
 
 	gtk_tree_model_get (model, iter,
-			    COL_NAME, &name,
+			    COL_REL_PATH, &path,
 			    -1);
-	if (!name) {
+	if (!path) {
 		return FALSE;
 	}
 
 	/* we never want to show these files */
-	if (strcmp (name, ".git") == 0 ||
-	    strcmp (name, ".gitignore") == 0 ||
-	    (!priv->show_all && file_list_ignore_file (model, iter, name))) {
+	if (g_str_has_suffix (path, ".git") ||
+	    g_str_has_suffix (path, ".gitignore") ||
+	    (!priv->show_all && file_list_get_ignore_file (model, iter, path))) {
 		retval = FALSE;
 	}
 
-	g_free (name);
+	g_free (path);
 
 	return retval;
 }
 
 static gboolean
-file_list_get_name_and_ignore_for_iter (GiggleFileList   *list,
+file_list_get_path_and_ignore_for_iter (GiggleFileList   *list,
 					GtkTreeIter      *iter,
-					gchar           **name,
+					gchar           **path,
 					GiggleGitIgnore **git_ignore)
 {
 	GiggleFileListPriv *priv;
@@ -565,18 +842,29 @@ file_list_get_name_and_ignore_for_iter (GiggleFileList   *list,
 	priv = GET_PRIV (list);
 
 	if (!gtk_tree_model_iter_parent (priv->filter_model, &parent, iter)) {
-		*name = NULL;
-		*git_ignore = NULL;
+		if (path) {
+			*path = NULL;
+		}
+
+		if (git_ignore) {
+			*git_ignore = NULL;
+		}
 
 		return FALSE;
 	}
 
-	gtk_tree_model_get (GTK_TREE_MODEL (priv->filter_model), iter,
-			    COL_NAME, name,
-			    -1);
-	gtk_tree_model_get (GTK_TREE_MODEL (priv->filter_model), &parent,
-			    COL_GIT_IGNORE, git_ignore,
-			    -1);
+	if (path) {
+		gtk_tree_model_get (GTK_TREE_MODEL (priv->filter_model), iter,
+				    COL_REL_PATH, path,
+				    -1);
+	}
+
+	if (git_ignore) {
+		gtk_tree_model_get (GTK_TREE_MODEL (priv->filter_model), &parent,
+				    COL_GIT_IGNORE, git_ignore,
+				    -1);
+	}
+
 	return TRUE;
 }
 
@@ -627,22 +915,64 @@ file_list_search_equal_func (GtkTreeModel *model,
 	return ret;
 }
 
+static gint
+file_list_compare_func (GtkTreeModel *model,
+			GtkTreeIter  *iter1,
+			GtkTreeIter  *iter2,
+			gpointer      user_data)
+{
+	GiggleGitIgnore *git_ignore1, *git_ignore2;
+	gchar           *name1, *name2;
+	gint             retval = 0;
+
+	gtk_tree_model_get (model, iter1,
+			    COL_GIT_IGNORE, &git_ignore1,
+			    COL_NAME, &name1,
+			    -1);
+	gtk_tree_model_get (model, iter2,
+			    COL_GIT_IGNORE, &git_ignore2,
+			    COL_NAME, &name2,
+			    -1);
+
+	if (git_ignore1 && !git_ignore2) {
+		retval = -1;
+	} else if (git_ignore2 && !git_ignore1) {
+		retval = 1;
+	} else {
+		retval = strcmp (name1, name2);
+	}
+
+	/* free stuff */
+	if (git_ignore1) {
+		g_object_unref (git_ignore1);
+	}
+
+	if (git_ignore2) {
+		g_object_unref (git_ignore2);
+	}
+
+	g_free (name1);
+	g_free (name2);
+
+	return retval;
+}
+
 static void
-file_list_add_file_foreach (GtkTreeModel *model,
-			    GtkTreePath  *path,
-			    GtkTreeIter  *iter,
-			    gpointer      data)
+file_list_ignore_file_foreach (GtkTreeModel *model,
+			       GtkTreePath  *path,
+			       GtkTreeIter  *iter,
+			       gpointer      data)
 {
 	GiggleGitIgnore *git_ignore;
 	gchar           *name;
 
-	if (!file_list_get_name_and_ignore_for_iter (GIGGLE_FILE_LIST (data),
+	if (!file_list_get_path_and_ignore_for_iter (GIGGLE_FILE_LIST (data),
 						     iter, &name, &git_ignore)) {
 		return;
 	}
 
 	if (git_ignore) {
-		giggle_git_ignore_add_glob (git_ignore, name);
+		giggle_git_ignore_add_glob_for_path (git_ignore, name);
 		g_object_unref (git_ignore);
 	}
 
@@ -650,8 +980,8 @@ file_list_add_file_foreach (GtkTreeModel *model,
 }
 
 static void
-file_list_add_file (GtkWidget      *widget,
-		    GiggleFileList *list)
+file_list_ignore_file (GtkAction      *action,
+		       GiggleFileList *list)
 {
 	GiggleFileListPriv *priv;
 	GtkTreeSelection   *selection;
@@ -659,28 +989,28 @@ file_list_add_file (GtkWidget      *widget,
 	priv = GET_PRIV (list);
 
 	selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (list));
-	gtk_tree_selection_selected_foreach (selection, file_list_add_file_foreach, list);
+	gtk_tree_selection_selected_foreach (selection, file_list_ignore_file_foreach, list);
 	gtk_tree_model_filter_refilter (GTK_TREE_MODEL_FILTER (priv->filter_model));
 }
 
 static void
-file_list_remove_file_foreach (GtkTreeModel *model,
-			       GtkTreePath  *path,
-			       GtkTreeIter  *iter,
-			       gpointer      data)
+file_list_unignore_file_foreach (GtkTreeModel *model,
+				 GtkTreePath  *tree_path,
+				 GtkTreeIter  *iter,
+				 gpointer      data)
 {
 	GiggleFileList  *list;
 	GiggleGitIgnore *git_ignore;
-	gchar           *name;
+	gchar           *path;
 
 	list = GIGGLE_FILE_LIST (data);
 
-	if (!file_list_get_name_and_ignore_for_iter (list, iter, &name, &git_ignore)) {
+	if (!file_list_get_path_and_ignore_for_iter (list, iter, &path, &git_ignore)) {
 		return;
 	}
 
 	if (git_ignore) {
-		if (!giggle_git_ignore_remove_glob_for_name (git_ignore, name, TRUE)) {
+		if (!giggle_git_ignore_remove_glob_for_path (git_ignore, path, TRUE)) {
 			GtkWidget *dialog, *toplevel;
 
 			toplevel = gtk_widget_get_toplevel (GTK_WIDGET (list));
@@ -695,7 +1025,7 @@ file_list_remove_file_foreach (GtkTreeModel *model,
 								    "that may be hiding other files, delete it?"));
 
 			if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_YES) {
-				giggle_git_ignore_remove_glob_for_name (git_ignore, name, FALSE);
+				giggle_git_ignore_remove_glob_for_path (git_ignore, path, FALSE);
 			}
 
 			gtk_widget_destroy (dialog);
@@ -704,12 +1034,12 @@ file_list_remove_file_foreach (GtkTreeModel *model,
 		g_object_unref (git_ignore);
 	}
 
-	g_free (name);
+	g_free (path);
 }
 
 static void
-file_list_remove_file (GtkWidget      *widget,
-		       GiggleFileList *list)
+file_list_unignore_file (GtkAction      *action,
+			 GiggleFileList *list)
 {
 	GiggleFileListPriv *priv;
 	GtkTreeSelection   *selection;
@@ -717,19 +1047,19 @@ file_list_remove_file (GtkWidget      *widget,
 	priv = GET_PRIV (list);
 
 	selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (list));
-	gtk_tree_selection_selected_foreach (selection, file_list_remove_file_foreach, list);
+	gtk_tree_selection_selected_foreach (selection, file_list_unignore_file_foreach, list);
 	gtk_tree_model_filter_refilter (GTK_TREE_MODEL_FILTER (priv->filter_model));
 }
 
 static void
-file_list_toggle_show_all (GtkWidget      *widget,
+file_list_toggle_show_all (GtkAction      *action,
 			   GiggleFileList *list)
 {
 	GiggleFileListPriv *priv;
 	gboolean active;
 
 	priv = GET_PRIV (list);
-	active = gtk_toggle_action_get_active (GTK_TOGGLE_ACTION (widget));
+	active = gtk_toggle_action_get_active (GTK_TOGGLE_ACTION (action));
 	giggle_file_list_set_show_all (list, active);
 }
 
@@ -743,7 +1073,7 @@ file_list_cell_data_sensitive_func (GtkCellLayout   *layout,
 	GiggleFileList     *list;
 	GiggleFileListPriv *priv;
 	GiggleGitIgnore    *git_ignore = NULL;
-	gchar              *name = NULL;
+	gchar              *path = NULL;
 	gboolean            value = TRUE;
 	GtkTreeIter         parent;
 	GtkStateType        state;
@@ -752,27 +1082,38 @@ file_list_cell_data_sensitive_func (GtkCellLayout   *layout,
 	list = GIGGLE_FILE_LIST (data);
 	priv = GET_PRIV (list);
 
+	/* we want highlighted background too */
+	file_list_cell_data_background_func (layout, renderer, tree_model, iter, data);
+
 	if (priv->show_all) {
-		if (file_list_get_name_and_ignore_for_iter (list, iter, &name, &git_ignore)) {
-			value = ! giggle_git_ignore_name_matches (git_ignore, name);
+		if (file_list_get_path_and_ignore_for_iter (list, iter, &path, &git_ignore)) {
+			value = ! giggle_git_ignore_path_matches (git_ignore, path);
 		} else {
 			/* we don't want the project root being set insensitive */
 			value = ! gtk_tree_model_iter_parent (tree_model, &parent, iter);
 		}
 	}
 
+	if (value) {
+		/* check whether the file is managed by git */
+		gtk_tree_model_get (tree_model, iter,
+				    COL_MANAGED, &value,
+				    -1);
+		value ^= 1;
+	}
+
 	if (GTK_IS_CELL_RENDERER_TEXT (renderer)) {
 		state = (value) ? GTK_STATE_NORMAL : GTK_STATE_INSENSITIVE;
 		color = GTK_WIDGET (list)->style->text [state];
-		g_object_set (G_OBJECT (renderer), "foreground-gdk", &color, NULL);
+		g_object_set (renderer, "foreground-gdk", &color, NULL);
 	} else {
-		g_object_set (G_OBJECT (renderer), "sensitive", value, NULL);
+		g_object_set (renderer, "sensitive", value, NULL);
 	}
 
 	if (git_ignore) {
 		g_object_unref (git_ignore);
 	}
-	g_free (name);
+	g_free (path);
 }
 
 static void
@@ -831,132 +1172,6 @@ giggle_file_list_set_show_all (GiggleFileList *list,
 	g_object_notify (G_OBJECT (list), "show-all");
 }
 
-static gint
-file_list_compare_prefix (gconstpointer a,
-			  gconstpointer b)
-{
-	return ! g_str_has_prefix (a, b);
-}
-
-static void
-file_list_update_highlight (GiggleFileList *file_list,
-			    GtkTreeIter    *parent,
-			    const gchar    *parent_path,
-			    GList          *files)
-{
-	GiggleFileListPriv *priv;
-	GtkTreeIter         iter;
-	gboolean            valid;
-	GiggleGitIgnore    *git_ignore;
-	gchar              *name, *path;
-	gboolean            highlight;
-
-	priv = GET_PRIV (file_list);
-
-	if (parent) {
-		valid = gtk_tree_model_iter_children (GTK_TREE_MODEL (priv->store),
-						      &iter, parent);
-	} else {
-		valid = gtk_tree_model_get_iter_first (GTK_TREE_MODEL (priv->store), &iter);
-	}
-
-	while (valid) {
-		gtk_tree_model_get (GTK_TREE_MODEL (priv->store), &iter,
-				    COL_NAME, &name,
-				    COL_GIT_IGNORE, &git_ignore,
-				    -1);
-
-		if (parent_path) {
-			path = g_build_filename (parent_path, name, NULL);
-			highlight = (g_list_find_custom (files, path, (GCompareFunc) file_list_compare_prefix) != NULL);
-		} else {
-			/* we don't want the project basename included */
-			path = g_strdup ("");
-			highlight = FALSE;
-		}
-
-		gtk_tree_store_set (priv->store, &iter,
-				    COL_HIGHLIGHT, highlight,
-				    -1);
-
-		if (git_ignore) {
-			/* it's a directory */
-			file_list_update_highlight (file_list, &iter, path, files);
-			g_object_unref (git_ignore);
-		}
-
-		g_free (path);
-		g_free (name);
-		valid = gtk_tree_model_iter_next (GTK_TREE_MODEL (priv->store), &iter);
-	}
-}
-
-static void
-file_list_job_callback (GiggleGit *git,
-			GiggleJob *job,
-			GError    *error,
-			gpointer   user_data)
-{
-	GiggleFileList     *list;
-	GiggleFileListPriv *priv;
-	GList              *files;
-
-	list = GIGGLE_FILE_LIST (user_data);
-	priv = GET_PRIV (list);
-
-	if (error) {
-		GtkWidget *dialog;
-
-		dialog = gtk_message_dialog_new (GTK_WINDOW (gtk_widget_get_toplevel (GTK_WIDGET (list))),
-						 GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
-						 GTK_MESSAGE_ERROR,
-						 GTK_BUTTONS_OK,
-						 _("An error ocurred when retrieving different files list:\n%s"),
-						 error->message);
-
-		gtk_dialog_run (GTK_DIALOG (dialog));
-		gtk_widget_destroy (dialog);
-	} else {
-		files = giggle_git_diff_tree_get_files (GIGGLE_GIT_DIFF_TREE (priv->job));
-		file_list_update_highlight (list, NULL, NULL, files);
-	}
-
-	g_object_unref (priv->job);
-	priv->job = NULL;
-}
-
-void
-giggle_file_list_highlight_revisions (GiggleFileList *list,
-				      GiggleRevision *from,
-				      GiggleRevision *to)
-{
-	GiggleFileListPriv *priv;
-
-	g_return_if_fail (GIGGLE_IS_FILE_LIST (list));
-	g_return_if_fail (GIGGLE_IS_REVISION (from));
-	g_return_if_fail (GIGGLE_IS_REVISION (to));
-
-	priv = GET_PRIV (list);
-
-	/* FIXME: should set properties */
-
-	/* clear highlights */
-	file_list_update_highlight (list, NULL, NULL, NULL);
-
-	if (priv->job) {
-		giggle_git_cancel_job (priv->git, priv->job);
-		g_object_unref (priv->job);
-		priv->job = NULL;
-	}
-
-	priv->job = giggle_git_diff_tree_new (from, to);
-
-	giggle_git_run_job (priv->git,
-			    priv->job,
-			    file_list_job_callback,
-			    list);
-}
-
 GList *
 giggle_file_list_get_selection (GiggleFileList *list)
 {
@@ -1037,3 +1252,125 @@ giggle_file_list_set_compact_mode (GiggleFileList *list,
 		g_object_notify (G_OBJECT (list), "compact-mode");
 	}
 }
+
+static gint
+file_list_compare_prefix (gconstpointer a,
+			  gconstpointer b)
+{
+	return ! g_str_has_prefix (a, b);
+}
+
+static void
+file_list_update_highlight (GiggleFileList *file_list,
+			    GtkTreeIter    *parent,
+			    const gchar    *parent_path,
+			    GList          *files)
+{
+	GiggleFileListPriv *priv;
+	GtkTreeIter         iter;
+	gboolean            valid;
+	GiggleGitIgnore    *git_ignore;
+	gchar              *rel_path, *name;
+	gboolean            highlight;
+
+	priv = GET_PRIV (file_list);
+
+	if (parent) {
+		valid = gtk_tree_model_iter_children (GTK_TREE_MODEL (priv->store),
+						      &iter, parent);
+	} else {
+		valid = gtk_tree_model_get_iter_first (GTK_TREE_MODEL (priv->store), &iter);
+	}
+
+	while (valid) {
+		gtk_tree_model_get (GTK_TREE_MODEL (priv->store), &iter,
+				    COL_NAME, &name,
+				    COL_REL_PATH, &rel_path,
+				    COL_GIT_IGNORE, &git_ignore,
+				    -1);
+
+		if (rel_path && *rel_path) {
+			highlight = (g_list_find_custom (files, rel_path, (GCompareFunc) file_list_compare_prefix) != NULL);
+		} else {
+			/* we don't want the project basename included */
+			highlight = FALSE;
+		}
+
+		gtk_tree_store_set (priv->store, &iter,
+				    COL_HIGHLIGHT, highlight,
+				    -1);
+
+		if (git_ignore) {
+			/* it's a directory */
+			file_list_update_highlight (file_list, &iter, rel_path, files);
+			g_object_unref (git_ignore);
+		}
+
+		g_free (rel_path);
+		valid = gtk_tree_model_iter_next (GTK_TREE_MODEL (priv->store), &iter);
+	}
+}
+
+static void
+file_list_job_callback (GiggleGit *git,
+			GiggleJob *job,
+			GError    *error,
+			gpointer   user_data)
+{
+	GiggleFileList     *list;
+	GiggleFileListPriv *priv;
+	GList              *files;
+
+	list = GIGGLE_FILE_LIST (user_data);
+	priv = GET_PRIV (list);
+
+	if (error) {
+		GtkWidget *dialog;
+
+		dialog = gtk_message_dialog_new (GTK_WINDOW (gtk_widget_get_toplevel (GTK_WIDGET (list))),
+						 GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
+						 GTK_MESSAGE_ERROR,
+						 GTK_BUTTONS_OK,
+						 _("An error ocurred when retrieving different files list:\n%s"),
+						 error->message);
+
+		gtk_dialog_run (GTK_DIALOG (dialog));
+		gtk_widget_destroy (dialog);
+	} else {
+		files = giggle_git_diff_tree_get_files (GIGGLE_GIT_DIFF_TREE (priv->job));
+		file_list_update_highlight (list, NULL, NULL, files);
+	}
+
+	g_object_unref (priv->job);
+	priv->job = NULL;
+}
+
+void
+giggle_file_list_highlight_revisions (GiggleFileList *list,
+				      GiggleRevision *from,
+				      GiggleRevision *to)
+{
+	GiggleFileListPriv *priv;
+
+	g_return_if_fail (GIGGLE_IS_FILE_LIST (list));
+	g_return_if_fail (GIGGLE_IS_REVISION (from));
+	g_return_if_fail (GIGGLE_IS_REVISION (to));
+
+	priv = GET_PRIV (list);
+
+	/* clear highlights */
+	file_list_update_highlight (list, NULL, NULL, NULL);
+
+	if (priv->job) {
+		giggle_git_cancel_job (priv->git, priv->job);
+		g_object_unref (priv->job);
+		priv->job = NULL;
+	}
+
+	priv->job = giggle_git_diff_tree_new (from, to);
+
+	giggle_git_run_job (priv->git,
+			    priv->job,
+			    file_list_job_callback,
+			    list);
+}
