@@ -34,11 +34,14 @@
 #include "giggle-diff-view.h"
 #include "giggle-diff-tree-view.h"
 #include "giggle-searchable.h"
+#include "giggle-history.h"
 
 typedef struct GiggleViewHistoryPriv GiggleViewHistoryPriv;
 
 struct GiggleViewHistoryPriv {
 	GtkWidget *file_list;
+	GtkWidget *file_list_sw;
+
 	GtkWidget *revision_list;
 	GtkWidget *revision_view;
 	GtkWidget *diff_view;
@@ -54,11 +57,18 @@ struct GiggleViewHistoryPriv {
 
 	GiggleGit *git;
 	GiggleJob *job;
+
+	GList     *history; /* reversed list of history elems */
+	GList     *current_history_elem;
+
+	guint     compact_mode : 1;
 };
 
 static void     view_history_finalize              (GObject *object);
 
 static void     giggle_view_history_searchable_init             (GiggleSearchableIface *iface);
+static void     giggle_view_history_history_init                (GiggleHistoryIface    *iface);
+
 static void     view_history_revision_list_selection_changed_cb (GiggleRevisionList *list,
 								 GiggleRevision     *revision1,
 								 GiggleRevision     *revision2,
@@ -69,14 +79,27 @@ static gboolean view_history_revision_list_key_press_cb         (GiggleRevisionL
 
 static gboolean view_history_search                             (GiggleSearchable      *searchable,
 								 const gchar           *search_term,
-								 GiggleSearchDirection  direction);
+								 GiggleSearchDirection  direction,
+								 gboolean               full_search);
+static void     view_history_cancel_search                      (GiggleSearchable      *searchable);
 
 static void     view_history_update_revisions                   (GiggleViewHistory  *view);
 
+static void     view_history_path_selected                      (GiggleDiffTreeView *diff_tree_view,
+								 const gchar        *path,
+								 GiggleViewHistory  *view);
+
+static void     view_history_go_back                            (GiggleHistory      *history);
+static gboolean view_history_can_go_back                        (GiggleHistory      *history);
+static void     view_history_go_forward                         (GiggleHistory      *history);
+static gboolean view_history_can_go_forward                     (GiggleHistory      *history);
+
 
 G_DEFINE_TYPE_WITH_CODE (GiggleViewHistory, giggle_view_history, GIGGLE_TYPE_VIEW,
 			 G_IMPLEMENT_INTERFACE (GIGGLE_TYPE_SEARCHABLE,
-						giggle_view_history_searchable_init))
+						giggle_view_history_searchable_init)
+			 G_IMPLEMENT_INTERFACE (GIGGLE_TYPE_HISTORY,
+						giggle_view_history_history_init))
 
 #define GET_PRIV(obj) (G_TYPE_INSTANCE_GET_PRIVATE ((obj), GIGGLE_TYPE_VIEW_HISTORY, GiggleViewHistoryPriv))
 
@@ -100,6 +123,16 @@ static void
 giggle_view_history_searchable_init (GiggleSearchableIface *iface)
 {
 	iface->search = view_history_search;
+	iface->cancel = view_history_cancel_search;
+}
+
+static void
+giggle_view_history_history_init (GiggleHistoryIface *iface)
+{
+	iface->go_back = view_history_go_back;
+	iface->can_go_back = view_history_can_go_back;
+	iface->go_forward = view_history_go_forward;
+	iface->can_go_forward = view_history_can_go_forward;
 }
 
 static void
@@ -111,6 +144,8 @@ giggle_view_history_init (GiggleViewHistory *view)
 
 	priv = GET_PRIV (view);
 
+	priv->compact_mode = FALSE;
+
 	gtk_widget_push_composite_child ();
 
 	priv->main_hpaned = gtk_hpaned_new ();
@@ -142,21 +177,27 @@ giggle_view_history_init (GiggleViewHistory *view)
 	gtk_paned_pack2 (GTK_PANED (priv->revision_hpaned), scrolled_window, FALSE, FALSE);
 	gtk_widget_show_all (scrolled_window);
 
+	g_signal_connect (priv->diff_tree_view, "path-selected",
+			  G_CALLBACK (view_history_path_selected), view);
+
 	vbox = gtk_vbox_new (FALSE, 0);
 	gtk_widget_show (vbox);
 	gtk_paned_pack1 (GTK_PANED (priv->revision_hpaned), vbox, TRUE, FALSE);
 
 	/* file view */
-	scrolled_window = gtk_scrolled_window_new (NULL, NULL);
-	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolled_window),
+	priv->file_list_sw = gtk_scrolled_window_new (NULL, NULL);
+	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (priv->file_list_sw),
 					GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
-	gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (scrolled_window), GTK_SHADOW_IN);
+	gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (priv->file_list_sw), GTK_SHADOW_IN);
 
 	priv->file_list = giggle_file_list_new ();
-	gtk_container_add (GTK_CONTAINER (scrolled_window), priv->file_list);
-	gtk_widget_show_all (scrolled_window);
+	gtk_container_add (GTK_CONTAINER (priv->file_list_sw), priv->file_list);
+	gtk_widget_show_all (priv->file_list_sw);
 
-	gtk_paned_pack1 (GTK_PANED (priv->main_hpaned), scrolled_window, FALSE, FALSE);
+	g_signal_connect (priv->file_list, "path-selected",
+			  G_CALLBACK (view_history_path_selected), view);
+
+	gtk_paned_pack1 (GTK_PANED (priv->main_hpaned), priv->file_list_sw, FALSE, FALSE);
 
 	/* revisions list */
 	scrolled_window = gtk_scrolled_window_new (NULL, NULL);
@@ -165,9 +206,9 @@ giggle_view_history_init (GiggleViewHistory *view)
 	gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (scrolled_window), GTK_SHADOW_IN);
 
 	priv->revision_list = giggle_revision_list_new ();
-	g_signal_connect (G_OBJECT (priv->revision_list), "selection-changed",
+	g_signal_connect (priv->revision_list, "selection-changed",
 			  G_CALLBACK (view_history_revision_list_selection_changed_cb), view);
-	g_signal_connect (G_OBJECT (priv->revision_list), "key-press-event",
+	g_signal_connect (priv->revision_list, "key-press-event",
 			  G_CALLBACK (view_history_revision_list_key_press_cb), view);
 
 	gtk_container_add (GTK_CONTAINER (scrolled_window), priv->revision_list);
@@ -205,7 +246,7 @@ giggle_view_history_init (GiggleViewHistory *view)
 
 	/* git interaction */
 	priv->git = giggle_git_get ();
-	g_signal_connect_swapped (G_OBJECT (priv->git), "notify::git-dir",
+	g_signal_connect_swapped (priv->git, "notify::git-dir",
 				  G_CALLBACK (view_history_update_revisions), view);
 	view_history_update_revisions (view);
 }
@@ -223,7 +264,12 @@ view_history_finalize (GObject *object)
 		priv->job = NULL;
 	}
 
+	g_list_foreach (priv->history, (GFunc) g_free, NULL);
+	g_list_free (priv->history);
+
 	g_object_unref (priv->git);
+
+	G_OBJECT_CLASS (giggle_view_history_parent_class)->finalize (object);
 }
 
 static void
@@ -233,17 +279,25 @@ view_history_revision_list_selection_changed_cb (GiggleRevisionList *list,
 						 GiggleViewHistory  *view)
 {
 	GiggleViewHistoryPriv *priv;
+	GList                 *files;
 
 	priv = GET_PRIV (view);
+	files = NULL;
 
 	giggle_revision_view_set_revision (
 		GIGGLE_REVISION_VIEW (priv->revision_view), revision1);
 
 	if (revision1 && revision2) {
+		if (priv->current_history_elem) {
+			files = g_list_prepend (NULL, g_strdup ((gchar *) priv->current_history_elem->data));
+		}
+
 		giggle_diff_view_set_revisions (GIGGLE_DIFF_VIEW (priv->diff_view),
-						revision1, revision2, NULL);
+						revision1, revision2, files);
 		giggle_diff_tree_view_set_revisions (GIGGLE_DIFF_TREE_VIEW (priv->diff_tree_view),
 						     revision1, revision2);
+		giggle_file_list_highlight_revisions (GIGGLE_FILE_LIST (priv->file_list),
+						revision1, revision2);
 	}
 }
 
@@ -270,7 +324,7 @@ view_history_revision_list_key_press_cb (GiggleRevisionList *list,
 
 		value = CLAMP (value, adj->lower, adj->upper - adj->page_size);
 
-		g_object_set (G_OBJECT (adj), "value", value, NULL);
+		g_object_set (adj, "value", value, NULL);
 
 		return TRUE;
 	}
@@ -281,28 +335,42 @@ view_history_revision_list_key_press_cb (GiggleRevisionList *list,
 static gboolean
 view_history_search (GiggleSearchable      *searchable,
 		     const gchar           *search_term,
-		     GiggleSearchDirection  direction)
+		     GiggleSearchDirection  direction,
+		     gboolean               full_search)
 {
 	GiggleViewHistoryPriv *priv;
 
 	priv = GET_PRIV (searchable);
 
 	if (!giggle_searchable_search (GIGGLE_SEARCHABLE (priv->revision_list),
-				       search_term, direction)) {
+				       search_term, direction, full_search)) {
 		return FALSE;
 	}
 
 	if (giggle_searchable_search (GIGGLE_SEARCHABLE (priv->revision_view),
-				      search_term, direction)) {
+				      search_term, direction, full_search)) {
 		/* search term is contained in the
 		 * revision description, expand it
 		 */
 		gtk_expander_set_expanded (GTK_EXPANDER (priv->revision_view_expander), TRUE);
+	} else if (giggle_searchable_search (GIGGLE_SEARCHABLE (priv->diff_view),
+					     search_term, direction, full_search)) {
+		gtk_expander_set_expanded (GTK_EXPANDER (priv->diff_view_expander), TRUE);
 	}
 
 	return TRUE;
 }
 
+static void
+view_history_cancel_search (GiggleSearchable *searchable)
+{
+	GiggleViewHistoryPriv *priv;
+
+	priv = GET_PRIV (searchable);
+
+	giggle_searchable_cancel (GIGGLE_SEARCHABLE (priv->revision_list));
+}
+
 typedef void (AddRefFunc) (GiggleRevision*, GiggleRef*);
 
 static gboolean
@@ -447,6 +515,8 @@ view_history_get_revisions_cb (GiggleGit    *git,
 				    priv->job,
 				    view_history_get_branches_cb,
 				    view);
+
+		giggle_history_changed (GIGGLE_HISTORY (view));
 	}
 }
 
@@ -454,6 +524,7 @@ static void
 view_history_update_revisions (GiggleViewHistory  *view)
 {
 	GiggleViewHistoryPriv *priv;
+	GList                 *files;
 
 	priv = GET_PRIV (view);
 
@@ -465,7 +536,14 @@ view_history_update_revisions (GiggleViewHistory  *view)
 		priv->job = NULL;
 	}
 
-	priv->job = giggle_git_revisions_new ();
+	if (priv->current_history_elem &&
+	    priv->current_history_elem->data) {
+		/* we just want one file */
+		files = g_list_prepend (NULL, g_strdup ((gchar *) priv->current_history_elem->data));
+		priv->job = giggle_git_revisions_new_for_files (files);
+	} else {
+		priv->job = giggle_git_revisions_new ();
+	}
 
 	giggle_git_run_job (priv->git,
 			    priv->job,
@@ -473,6 +551,38 @@ view_history_update_revisions (GiggleViewHistory  *view)
 			    view);
 }
 
+static void
+view_history_path_selected (GiggleDiffTreeView *diff_tree_view,
+			    const gchar        *path,
+			    GiggleViewHistory  *view)
+{
+	GiggleViewHistoryPriv *priv;
+	GList                 *list = NULL;
+
+	priv = GET_PRIV (view);
+
+	if (priv->current_history_elem) {
+		list = priv->current_history_elem;
+
+		if (list->prev) {
+			/* unlink from the first elements */
+			list->prev->next = NULL;
+			list->prev = NULL;
+
+			g_list_foreach (priv->history, (GFunc) g_free, NULL);
+			g_list_free (priv->history);
+		}
+	} else {
+		g_list_foreach (priv->history, (GFunc) g_free, NULL);
+		g_list_free (priv->history);
+	}
+
+	list = g_list_prepend (list, g_strdup (path));
+	priv->history = priv->current_history_elem = list;
+
+	view_history_update_revisions (view);
+}
+
 GtkWidget *
 giggle_view_history_new (void)
 {
@@ -491,4 +601,101 @@ giggle_view_history_set_compact_mode (GiggleViewHistory *view,
 
 	giggle_file_list_set_compact_mode (GIGGLE_FILE_LIST (priv->file_list), compact_mode);
 	giggle_revision_list_set_compact_mode (GIGGLE_REVISION_LIST (priv->revision_list), compact_mode);
+	giggle_diff_view_set_compact_mode (GIGGLE_DIFF_VIEW (priv->diff_view), compact_mode);
+	priv->compact_mode = compact_mode;
+}
+
+gboolean
+giggle_view_history_get_compact_mode  (GiggleViewHistory *view)
+{
+	GiggleViewHistoryPriv *priv;
+
+	g_return_val_if_fail (GIGGLE_IS_VIEW_HISTORY (view), FALSE);
+
+	priv = GET_PRIV (view);
+	return priv->compact_mode;
+}
+
+void
+giggle_view_history_set_file_list_visible (GiggleViewHistory *view,
+					   gboolean           visible)
+{
+	GiggleViewHistoryPriv *priv;
+
+	g_return_if_fail (GIGGLE_IS_VIEW_HISTORY (view));
+
+	priv = GET_PRIV (view);
+
+	g_object_set (priv->file_list_sw, "visible", visible, NULL);
+}
+
+gboolean
+giggle_view_history_get_file_list_visible (GiggleViewHistory *view)
+{
+	GiggleViewHistoryPriv *priv;
+
+	g_return_val_if_fail (GIGGLE_IS_VIEW_HISTORY (view), FALSE);
+
+	priv = GET_PRIV (view);
+
+	return GTK_WIDGET_VISIBLE (priv->file_list_sw);
+}
+
+static void
+view_history_go_back (GiggleHistory *history)
+{
+	GiggleViewHistoryPriv *priv;
+
+	priv = GET_PRIV (history);
+
+	if (priv->current_history_elem) {
+		priv->current_history_elem = priv->current_history_elem->next;
+		giggle_history_changed (history);
+		view_history_update_revisions (GIGGLE_VIEW_HISTORY (history));
+	}
+}
+
+static gboolean
+view_history_can_go_back (GiggleHistory *history)
+{
+	GiggleViewHistoryPriv *priv;
+
+	priv = GET_PRIV (history);
+
+	return (priv->current_history_elem != NULL ||
+		g_list_length (priv->history) > 2);
+}
+
+static void
+view_history_go_forward (GiggleHistory *history)
+{
+	GiggleViewHistoryPriv *priv;
+	gboolean               changed = FALSE;
+
+	priv = GET_PRIV (history);
+
+	if (!priv->current_history_elem) {
+		/* go to the last one in list (first in history) */
+		priv->current_history_elem = g_list_last (priv->history);
+		changed = TRUE;
+	} else if (priv->current_history_elem != priv->history) {
+		priv->current_history_elem = priv->current_history_elem->prev;
+		changed = TRUE;
+	}
+
+	if (changed) {
+		view_history_update_revisions (GIGGLE_VIEW_HISTORY (history));
+		giggle_history_changed (history);
+	}
+}
+
+static gboolean
+view_history_can_go_forward (GiggleHistory *history)
+{
+	GiggleViewHistoryPriv *priv;
+
+	priv = GET_PRIV (history);
+
+	return (priv->current_history_elem != priv->history ||
+		g_list_length (priv->history) > 2);
 }
